/*
ggroup -- View and manipulate global groups on WIndows NT, WIndows 2000 and WIndows 2003 domains.
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  
	
*/
#include <windows.h>
#include <stdio.h>
#include <time.h> 
#include <lm.h>
#include <io.h>

 
#define RTN_OK 0
#define RTN_USAGE 4
#define RTN_NEVER 1
#define RTN_ERROR 8
#define GBUFSIZ 256000

// #define TRACE
 
void DisplayErrorText(DWORD dwLastError);
	FILE *otfi;
	char ErrorMessage[1024];
	int verbose = 0;
	int quiet = 0;
	int delimited = 0;
	char DomainName[64];
	int fullname = 0, members = 1, list = 0, terse = 0;
	char aord, addgroup = 0, delgroup = 0;
	wchar_t wComment[1024];

   //
   // Unicode entry point and argv
   //

#define NAMELEN_I 64

int GetFullName(LPWSTR wComputerName, LPWSTR wUserName, char **FullName) {
	DWORD nas;
    USER_INFO_2 *ui2;
	int namelen = NAMELEN_I;
	*FullName = malloc(namelen);
	nas = NetUserGetInfo(wComputerName,wUserName,2,(LPBYTE *) &ui2);
	if (nas == NERR_Success) {
		if (ui2->usri2_flags & UF_NORMAL_ACCOUNT) {
			// Get actual name length.
			// if *FullName is too small, then realloc to namelen+1
			namelen = wcslen(ui2->usri2_full_name);
			if (namelen >= NAMELEN_I)
				*FullName = realloc(*FullName,namelen+1);
			sprintf(*FullName,"%ls",ui2->usri2_full_name);
//			if (strlen(FullName) > 24) 
//			strcpy(FullName+20," ...");
//			strcat(UserName,FullName);
		}
		else if (ui2->usri2_flags & UF_TEMP_DUPLICATE_ACCOUNT)
			strcpy(*FullName,"[Local Account]");
		else if (ui2->usri2_flags & UF_WORKSTATION_TRUST_ACCOUNT)
			strcpy(*FullName,"[Machine Account]");
		else if (ui2->usri2_flags & UF_SERVER_TRUST_ACCOUNT)
			strcpy(*FullName,"[Domain Server Account]");
		else if (ui2->usri2_flags & UF_INTERDOMAIN_TRUST_ACCOUNT)
			strcpy(*FullName,"[Permit Trust Account]");
		else 
			strcpy(*FullName,"[Unknown Account Type]");
		NetApiBufferFree((PVOID) ui2);
	}
	else  {
		// Have an error, free the buffer
		free(*FullName);
		fprintf(stderr,"Error retrieving user information for: %ls\n", wUserName);								
		fprintf(otfi,"%-32ls Error Retrieving user information.\n",wUserName);
		DisplayErrorText( nas );
		return RTN_ERROR;
	}
	// Good return, buffer will be freed by caller
	return 0;
}

#define USER_COUNT 64000
//#define USER_COUNT 64
#define USERNAMELEN 32
GROUP_USERS_INFO_0 groupmembers[USER_COUNT];
int groupmembercount;

void gather_users(
	FILE *otfi, int starting_member, int argc, wchar_t **argv) {
	
	int i,k,len;
	FILE *inlist;
	char temp[_MAX_PATH];

	// save the list of users to add or delete
	for (groupmembercount=0,i=starting_member; i < argc && groupmembercount < USER_COUNT; ++i) {
		sprintf(temp,"%ls",argv[i]);
		if (temp[0] == '@') {
			inlist = fopen(temp+1,"r");
			if (inlist == NULL) {
				fprintf(otfi,"Unable to open: %s\n",temp+1);
				fprintf(stderr,"Unable to open: %s\n",temp+1);
			}
			else {
				while(fgets(temp,USERNAMELEN,inlist) && groupmembercount < USER_COUNT) {
					len = strlen(temp);
					temp[len-1] = 0; // remove ending lf

					groupmembers[groupmembercount].grui0_name = (LPWSTR) malloc(len * sizeof(WCHAR));
					k=mbstowcs(groupmembers[groupmembercount].grui0_name,temp,len);
				    sprintf(temp,"%ls",groupmembers[groupmembercount].grui0_name);
					++(groupmembercount);
					// fprintf(otfi,"\t%s\n",temp);
				}
				fclose(inlist);
			}
		}
		else {
			// localgroup_members[groupmembercount].lgrmi3_domainandname = argv[i];
			len = wcslen(argv[i]) + 1;
			groupmembers[groupmembercount].grui0_name = (LPWSTR) malloc(len*sizeof(WCHAR));
			wcscpy(groupmembers[groupmembercount].grui0_name, argv[i]);
			sprintf(temp,"%ls",groupmembers[groupmembercount].grui0_name);
			++(groupmembercount);
		}
	}
}

int add_delete_users(FILE *otfi, LPWSTR wComputerName, 
					 LPWSTR wGroupName, const char addordel) {

	char member[_MAX_PATH];
	int i, rc=0;
    NET_API_STATUS  nas=0;
	

	// Print out list of users to add / delete
	fprintf(otfi,"PDC: %ls. Group: %ls. %s:\n",
	   wComputerName, wGroupName,
	   (toupper(addordel)=='A' ? "Adding" : 
	   (toupper(addordel)=='D' ? "Deleting" : 
	   (toupper(addordel)=='S' ? "Setting Members to" : "???"))));
	
	for (i=0; i < groupmembercount; ++i) {
		sprintf(member,"%ls:",groupmembers[i]);

		if (strchr("ADS",toupper(addordel)) == NULL) {
			fprintf(otfi,"%s: Skipped.\n",member);
			continue;
		}

		switch(toupper(addordel)) {
		case 'A': 
			nas = NetGroupAddUser(wComputerName,wGroupName,groupmembers[i].grui0_name); 
			break;
		case 'D':
			nas = NetGroupDelUser(wComputerName,wGroupName,groupmembers[i].grui0_name); 
			break;
		case 'S':
			// This loop prints out the list of users and sets i so that the outer
			// loop will also stop
			for (i=0; i < groupmembercount; ++i)
				fprintf(otfi,"%ls\n",groupmembers[i]);
			nas = NetGroupSetUsers(wComputerName,wGroupName,0,(LPBYTE)groupmembers,(DWORD)groupmembercount);
			member[0] = 0;
			break;
		}

		// check the status of the command
		if (nas == NERR_Success) {
			fprintf(otfi,"%s Success\n",member);
			rc = 0;
		}
		else {
			fprintf(stderr,"%s Failed.  ",member);
			DisplayErrorText( nas );
			fprintf(otfi, "%s Failed.  %s",member, ErrorMessage);
			rc = -1;
		}

	}

	for (i=0; i < groupmembercount; ++i)
		free(groupmembers[i].grui0_name);

	return rc;
}




int ListGlobalGroupMembers(LPWSTR wGroupName, LPWSTR wComputerName, LPWSTR wComment,
						   FILE *otfi, int verbose, int fullname, int members, 
						   int delimited, int quiet) {
// Parse out domain name and group name; list all members of the group

	   GROUP_USERS_INFO_0 *gi0 = NULL;
	   DWORD entries_read, total_entries, resume_handle;

   char UserName[256],  ComputerName[32];
   char groups[256], *FullName = NULL;
   unsigned i, usercount = 0, totalcount=-1;

   NET_API_STATUS  nas;


	sprintf(ComputerName,"%ls",wComputerName);

    // Get the type 0 info into pgm2

		entries_read = 0;
		total_entries= 1;
		*groups = 0;
		resume_handle = 0;
		// Read in all of the Local Groups
		// First read in without indirect flag
		while (entries_read < total_entries) {
			nas = NetGroupGetUsers(wComputerName,wGroupName,0,(LPBYTE *) &gi0,
				GBUFSIZ,&entries_read, &total_entries, &resume_handle);

			if (nas == NERR_Success || nas == ERROR_MORE_DATA) {
				if (totalcount == -1)
					totalcount = total_entries;
				// Found More
				for (i=0; i < entries_read; ++i) {
					if (quiet == 0)
						fprintf(stderr,"                              \rUser %d of %d",++usercount,totalcount);

					if (fullname) {
						if (GetFullName(wComputerName,gi0[i].grui0_name,&FullName) == RTN_ERROR)
							return RTN_ERROR;
					}
					else {
						FullName = malloc(10);
						FullName[0] = 0;
					}
					if (verbose) {
						sprintf(UserName,"%s\\%ls", DomainName, gi0[i].grui0_name);
						    // Get the type 3 info into pi3
						if (delimited) 
							fprintf(otfi,"%s\\%ls\t%s\t%s\n",DomainName,wGroupName,UserName,FullName);
						else
							fprintf(otfi,"%-38s %s\n",UserName,FullName);
					}
					else
						if (delimited)
							fprintf(otfi,"%s\t%ls\t%ls\t%s\n",DomainName, wGroupName, gi0[i].grui0_name,FullName);
						else
							fprintf(otfi,"%-20ls %s\n", gi0[i].grui0_name, FullName);
					free(FullName);
				}
			}
			else  {
				fprintf(stderr,"Error retrieving group users for Group: %ls in %ls\n", 
					wGroupName, wComputerName);
				DisplayErrorText( nas );
				fprintf(otfi,"Group:   %-32ls. %s\n",wGroupName,  ErrorMessage);
//				NetApiBufferFree((PVOID) gi0);
				return RTN_ERROR;
			}

		}	
		if (total_entries == 0 && delimited)
			if (verbose)
				fprintf(otfi,"%s\\%ls\t%s\\<EmptyGroup>\n",DomainName,wGroupName,DomainName);
			else
				fprintf(otfi,"%s\t%ls\t<EmptyGroup>\n",DomainName,wGroupName);

	NetApiBufferFree((PVOID) gi0);
	return 0;

}
 

int displayGroupInfo(WCHAR *wGroupName, WCHAR *wComment, WCHAR *wComputerName, 
					 int *firsttime, int delimited) {
	char gname[256];
	
#ifdef TRACE
	int xx;

#endif
//				wcscpy(wGroupName, gi1[localgroup].grpi1_name);
//				wcscpy(wComment, gi1[localgroup].grpi1_comment); 
#ifdef TRACE
	xx = RTN_ERROR;
	fprintf(stderr,"localgroup=%d, entriesread=%d, rc=%d RTN_ERROR=%d\n",
		localgroup, entriesread, rc, xx);
#endif

	// Display Group Name
	if (*firsttime) 
		*firsttime = 0;
	else 
		if (members && !delimited)
			fprintf(otfi,"\n");
	// Terse Mode -- only print group name
	if (fullname==0)
		*wComment = 0;

	if (verbose)
		sprintf(gname,"%s\\%ls",DomainName, wGroupName);
	else
		sprintf(gname,"%ls", wGroupName);
	
#ifdef TRACE
	fprintf(stderr,"members=%d, len(gname)=%d gname = %s\n",
		members,strlen(gname),gname);
#endif

	if (terse) {
		if (delimited)
			fprintf(otfi,"%s\t%s\t%ls\n",DomainName,gname,wComment);
		else
			fprintf(otfi,"%-38s %ls\n", gname, wComment);
		return NERR_Success;
	}
			
	if (delimited)
		if (members)
			if (verbose)
				fprintf(otfi,"%s\t%s\\<GroupName>\t%ls\n",gname,DomainName,wComment);
			else
				fprintf(otfi,"%s\t%s\t<GroupName>\t%ls\n",DomainName,gname,wComment);
		else
			;
	else
		fprintf(otfi,"*Global Group: %-38s %ls\n", gname, wComment);

	if (members) {
		if (ListGlobalGroupMembers(wGroupName,wComputerName,wComment,otfi,verbose,fullname,members,delimited,quiet)) {
			return RTN_ERROR;
		}
	}

	return NERR_Success;
}


void set_temp_path(char *tmp) {
	// set tmp to the path to temp files
	char *q;

	if ((q=getenv("tmp")) == NULL)
		if ((q = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,q);
	else
		strcpy(tmp,q);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");
}

void displayHelp(WCHAR *progname,char *tmp) {
	// Use Notepad to display help
	char cmd[1024];
	FILE *hlp;
	// progname is the name of the command
	// tmp is the path to use for temp files

	strcat(tmp,"ggroup.hlp");

	hlp = fopen(tmp,"w");
	
	
	fprintf(hlp,"View / Manipulate Global Groups Copyright London Computing, 1985-2004, (WIN32) %s.\n",__DATE__);
	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");


	fprintf(hlp,"GGROUP %s\n",__TIMESTAMP__);
    fprintf(hlp, "Usage:\n%ls {<group>|-List} {<servername>|<domainname>} [-m] [-t] [-v] [-f]\n",progname);
	fprintf(hlp, "-or-\n");
	fprintf(hlp, "%ls group {<servername>|<domainname>} {-Add | -Del | -Set} <mbr> [<mbr> ...]\n",progname);
	fprintf(hlp, "-or-\n");
	fprintf(hlp, "%ls group {<servername>|<domainname>} {[-Addgroup] [-Comment:\"Group Comment\"] | -Delgroup\n\n",progname);
	fprintf(hlp, "To Display all of the members of the specified global group, specify\n");
	fprintf(hlp, "<group>      is the global group name to Check\n");
	fprintf(hlp, "-List        display all global groups defined in this domain.\n");
	fprintf(hlp, "<servername> is the name of a PDC or BDC,must begin with \\\\\n");
	fprintf(hlp, "<domainname> is the name of an NT domain, must NOT begin with \\\\\n");
	fprintf(hlp, "-q           Quiet Mode, no extraneous messages.\n");
	fprintf(hlp, "-v           Display Users as: domain\\userid\n");
	fprintf(hlp, "-t           User only with -List, display only Group Names.\n");
	fprintf(hlp, "-m           Used only with -List, display group members with group names.\n");
	fprintf(hlp, "-f           Display Users FullName after userid, group comment after group name.\n");
	fprintf(hlp, "-d           Used only when listing members, displays entries tab delimited.\n");
	fprintf(hlp,"              groupname, domainname, username, fullname\n\n");
	fprintf(hlp, "-Addgroup    Creates a new group and sets the optional comment field.\n");
	fprintf(hlp, "-Comment     without -Addgroup, changes the group comment.\n");
	fprintf(hlp, "-Delgroup    Deletes the group.\n\n");
	fprintf(hlp, "To Add / Remove / Set members of the global group, specify\n");
	fprintf(hlp, "-Add or -Del or -Set.\n");
	fprintf(hlp, "<mbr> is either a username or if it begins with an @\n");
	fprintf(hlp, "<mbr> is the name of a file containing a list of usernames, 1 per line.\n\n");
	fprintf(hlp, "When displaying users in a group, a running count will be displayed on stderr.\n\n");
	fprintf(hlp, "The output of this command will be displayed in a NOTEPAD window unless it is\n");
	fprintf(hlp, "redirected to a file using > or >> or piped to another command using |.\n\n");
	fprintf(hlp, "%s%s%s",
	       "Returns:     0--OK\n",
		   "             4--This help displayed.\n",
		   "             8--Parameter Error.\n");
   	fclose(hlp);
	strcpy(cmd,"NOTEPAD.EXE ");
	strcat(cmd,tmp);

	WinExec(cmd,SW_SHOW);
}


void set_flag(WCHAR *arg) {
char options[256];

	sprintf(options,"%ls", arg);
	if (stricmp(options,"-Q") == 0)
		quiet = 1;
	else if (stricmp(options,"-V") == 0)
		verbose = 1;
	else if (stricmp(options,"-F") == 0) 
		fullname = 1;
	else if (stricmp(options,"-M") == 0)
		members = 1;
	else if (stricmp(options,"-T") == 0)
		terse = 1;
	else if (stricmp(options,"-Add") == 0 || 
			stricmp(options,"-Del")==0 ||
			stricmp(options,"-Set")==0) {
		aord = options[1];
	}
	else if (stricmp(options,"-AddGroup") == 0)
		addgroup = 1;
	else if (stricmp(options,"-Delgroup") == 0)
		delgroup = 1;
	else if (strnicmp(options,"-Comment:",9) == 0)
		wcscpy(wComment,arg+9);
	else if (strnicmp(options,"-C:",3) == 0)
		wcscpy(wComment,arg+3);
	else if (stricmp(options,"-D") == 0)
		delimited = 1;
	else {
		fprintf(stderr,"%ls -- Invalid.\n",arg);
		exit(8);
	}
}

int SetComputerNamefromDomainName(WCHAR *arg,WCHAR *wComputerName, WCHAR *WCN) { 
	WCHAR *wCN;
	// exit if bad domain name
	DWORD nas;

	// Check to see if arg 2 is a domain name or a computername
	if (arg[0] != L'\\' || arg[1] != L'\\') 
		nas = NetGetDCName(NULL,WCN,(LPBYTE *)&wCN);
	else
		nas = NERR_Success;

#ifdef TRACE
	fprintf(stderr,"nas=%d wComputerName=%ls\n",nas,wComputerName);
#endif

	if (nas == NERR_Success) {
		wcscpy(wComputerName, wCN);
		NetApiBufferFree(wCN);
		return(0);
	}
	else {
		fprintf(stderr,"Unable to locate domain controller for %ls\n", wComputerName);
		DisplayErrorText( nas );
		fprintf(otfi,"%s\n",ErrorMessage);
		if (otfi != stdout)
			fclose(otfi);
		return(RTN_ERROR);
    }
	return(0);
}



void CheckShortHandComputerName(char *ComputerName,WCHAR *wComputerName, WCHAR *WCN) {
	strupr(ComputerName);
}



int __cdecl wmain(int argc, wchar_t *argv[] ) {

//	   int days, minutes, hours, seconds, remainder;
	   int stdio_handle, rc = 0, firsttime=1, i;

       LPWSTR       wComputerName = NULL; // default to local machine
	   LPWSTR		wGroupName;
	   WCHAR		WCN[64];

	   DWORD entriesread, totalentries, resumehandle;
	   GROUP_INFO_1 *gi1, groupinfo1[1];
	   GROUP_INFO_1002 groupinfo1002;
	   PNET_DISPLAY_GROUP pBuff, p;
	 
	   char ComputerName[32];
   char GroupName[64], groups[128];
	
   unsigned localgrouplen=GBUFSIZ;

   NET_API_STATUS  nas;
	char addordel = 0;
	char cmd[_MAX_PATH], tmp[_MAX_PATH];
	
	// Notepad will be used to either display the data or the help
	// Get the path to a temp directory here
	set_temp_path(tmp);  // set variable tmp to the path for TEMP files

	otfi = stdout;

	groups[0] = 0;
	wComment[0] = 0;
   
   if( argc < 3 ) {
		// not enough arguments
	   displayHelp(argv[0],tmp);
       return RTN_USAGE;
   }
 
   //
   // process command line arguments
   //


    wGroupName = argv[1];
	wcscpy(WCN,argv[2]);
    wComputerName = WCN;

	sprintf(ComputerName,"%ls",wComputerName);
	sprintf(DomainName,"%ls", wComputerName);
	sprintf(GroupName,"%ls", wGroupName);

	if (stricmp(GroupName,"-List") == 0 || stricmp(GroupName,"*")==0) {
		strcpy(GroupName,"-List");
		list = 1;
		members = 0;
#ifdef TRACE
		fprintf(stderr,"GroupName=%ls ComputerName=%ls List=1 Members=0\n",
			wGroupName,wComputerName);
#endif
	}
	aord=0;

	for (i=3; i < argc; ++i) {
		// Set flags
		set_flag(argv[i]);
		if (aord) {
			gather_users(otfi, i+1, argc, argv);
			break;
		}
	}



#ifdef TRACE
	fprintf(stderr,"verbose=%d fullname=%d members=%d terse=%d options=%s\n",
		verbose,fullname,members,terse,options);
#endif
	
	stdio_handle = _fileno(stdout);

	if(! _isatty(stdio_handle) )
		// stdout has been redirected, don't create a temp file for notepad
		otfi = stdout;
	else {
		if (list)
			strcat(tmp,"GlobalGroups");
		else
			strcat(tmp,GroupName);
		// skip past any leading \\ in the computername
		for (i=0;i < (int)strlen(ComputerName) && ComputerName[i]=='\\';++i);
		if (ComputerName[i] == '-')
			++i;				// Skip past - in -NE...
		strcat(tmp,"_");
		strcat(tmp,ComputerName+i);
		strcat(tmp,".txt");
		otfi = fopen(tmp,"w");
		if (otfi == NULL) {
			fprintf(stderr, "Unable to create %s\n",tmp);
			fprintf(stderr, "Output will be sent to STDOUT\n");
			otfi = stdout;
		}
	}

#ifdef TRACE
	fprintf(stderr,"OTFI = %8x\n", otfi);
#endif


	CheckShortHandComputerName(ComputerName,wComputerName,WCN);
	if (rc=SetComputerNamefromDomainName(argv[2],wComputerName,WCN))
		return rc; // exit if bad domain name


	sprintf(GroupName,"%ls",wGroupName);
	sprintf(ComputerName,"%ls",wComputerName);

	if (addgroup) {
		fprintf(otfi,"Create Group %ls in domain %s: ",wGroupName,DomainName);
		groupinfo1[0].grpi1_name = wGroupName;
		groupinfo1[0].grpi1_comment = wComment;
		nas = NetGroupAdd(wComputerName,1,(LPBYTE) groupinfo1+0,NULL);
		if (nas == NERR_Success)
			fprintf(otfi,"Success");
		else{
			fprintf(stderr,"Unable to create group %ls on %ls\n", wGroupName, wComputerName);
			DisplayErrorText( nas );
			fprintf(otfi,"Failed.  %s\n",ErrorMessage);
			if (otfi != stdout)
				fclose(otfi);
			return RTN_ERROR;
		}
	}
	else if (*wComment) {
		fprintf(otfi,"Set Group Comment to: %ls for group %ls in domain %s: ",wComment,wGroupName,DomainName);
		groupinfo1002.grpi1002_comment = wComment;
		nas = NetGroupSetInfo(wComputerName,wGroupName,1002,(LPBYTE) &groupinfo1002,NULL);
		if (nas == NERR_Success)
			fprintf(otfi,"Success");
		else{
			fprintf(stderr,"Unable to create group %ls on %ls\n", wGroupName, wComputerName);
			DisplayErrorText( nas );
			fprintf(otfi,"Failed.  %s\n",ErrorMessage);
			if (otfi != stdout)
				fclose(otfi);
			return RTN_ERROR;
		}
	}
	else if (delgroup) {
		fprintf(otfi,"Delete group %ls on %ls: ",wGroupName,wComputerName);
		nas = NetGroupDel(wComputerName,wGroupName);
		if (nas == NERR_Success)
			fprintf(otfi,"Success");
		else{
			fprintf(stderr,"Unable to delete group %ls on %ls\n", wGroupName, wComputerName);
			DisplayErrorText( nas );
			fprintf(otfi,"Failed.  %s\n",ErrorMessage);
			if (otfi != stdout)
				fclose(otfi);
			return RTN_ERROR;
		}
	}

	else if (aord) 
		add_delete_users(otfi,wComputerName,wGroupName,aord);
	else {
#ifdef TRACE
		fprintf(stderr,"Start ENUM\n");
#endif
		// If extracting a list of members, run enum to get an array of groups
		resumehandle=0;
		entriesread=0;
		nas = ERROR_MORE_DATA;
		totalentries=1;
		if (delimited)
			if (verbose)
				fprintf(otfi,"GroupName\tUserName\tFullName\n");
			else
				fprintf(otfi,"DomainName\tGroupName\tUserName\tFullName\n");

		if (!list) {
			*wComment=0;
			gi1 = &groupinfo1[0];
			nas = NetGroupGetInfo(wComputerName, wGroupName, 1, (LPBYTE *)&gi1);
			if (nas == NERR_Success) {
				rc=displayGroupInfo(gi1->grpi1_name, gi1->grpi1_comment, wComputerName, &firsttime, delimited);
				NetApiBufferFree(gi1);
			}
			else {
				fprintf(stderr,"Unable to display group membership for %ls\\%ls\n", 
					wComputerName,wGroupName);
				DisplayErrorText( nas );
				fprintf(otfi,"%s\n",ErrorMessage);
//				if (otfi != stdout)
//					fclose(otfi);
//				return RTN_ERROR;
			}
		}

		// Fall into this loop, but it doesn't matter 
		// nas is set to ERROR_MORE_DATA only if list is set.
		while (nas == ERROR_MORE_DATA) {
		    // Call the NetQueryDisplayInformation function;
			//   specify information level 3 (group account information).
			//
			nas = NetQueryDisplayInformation(wComputerName, 3, resumehandle, 
				1000, 0xFFFFFFFF, &entriesread, &pBuff);
#ifdef TRACE
			fprintf(stderr,"list=%d nas=%d entriesread=%d totalentries=%d\n",
				list,nas,entriesread,totalentries);
#endif
			// If call failed, print error message
			if (nas != NERR_Success  && nas != ERROR_MORE_DATA) {
				fprintf(stderr,"Unable to enumerate groups for %ls\n", wComputerName);
				DisplayErrorText( nas );
				fprintf(otfi,"%s\n",ErrorMessage);
				if (otfi != stdout)
					fclose(otfi);
				return RTN_ERROR;
			}

			// Display the group information (name, comment);
//			for (p=pBuff; entriesread-- && rc != RTN_ERROR; 
			for (p=pBuff; entriesread-- ; 
				++i,resumehandle=p->grpi3_next_index,p++) {
//				Display the Infor about the group			
				rc=displayGroupInfo(p->grpi3_name,p->grpi3_comment,wComputerName,&firsttime,delimited);
			}
			
			NetApiBufferFree(pBuff);

		}

	}

	if (otfi == stdout)
		;
	else {
		fclose(otfi);
		strcpy(cmd,"NOTEPAD.EXE ");
		strcat(cmd,tmp);
		WinExec(cmd,SW_SHOW);
	}

    return 0;
}
 
void DisplayErrorText(DWORD dwLastError) {
       HMODULE hModule = NULL; // default to system source
       LPSTR MessageBuffer;
       DWORD dwBufferLength;
       DWORD dwFormatFlags;
 
       dwFormatFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                       FORMAT_MESSAGE_IGNORE_INSERTS |
                       FORMAT_MESSAGE_FROM_SYSTEM ;
 
       //
       // if dwLastError is in the network range, load the message source
       //
       if(dwLastError >= NERR_BASE && dwLastError <= MAX_NERR) {
           hModule = LoadLibraryEx(
               TEXT("netmsg.dll"),
               NULL,
               LOAD_LIBRARY_AS_DATAFILE
               );
 
           if(hModule != NULL)
               dwFormatFlags |= FORMAT_MESSAGE_FROM_HMODULE;
       }
 
       //
       // call FormatMessage() to allow for message text to be acquired
       // from the system or the supplied module handle.
       //
       if(dwBufferLength = FormatMessageA(
           dwFormatFlags,
           hModule, // module to get message from (NULL == system)
           dwLastError,
           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
           (LPSTR) &MessageBuffer,
           0,
           NULL
           ))
       {
           DWORD dwBytesWritten;
 
           //
           // Output message string on stderr
           //
           WriteFile(
               GetStdHandle(STD_ERROR_HANDLE),
               MessageBuffer,
               dwBufferLength,
               &dwBytesWritten,
               NULL
               );
 
           //
           // free the buffer allocated by the system
           //
		   strcpy(ErrorMessage, MessageBuffer);
           LocalFree(MessageBuffer);
       }
 
       //
       // if you loaded a message source, unload it.
       //
       if(hModule != NULL)
           FreeLibrary(hModule);


}

